//! Provides archive related functionalities

use std::path::PathBuf;

use super::config::{Config, RELATIVE_CONFIG_FILE};
use super::error::Archive as ArchiveError;
use super::git;

/// Represents an archive
pub struct Archive {
    pub path: PathBuf,
    pub repo: git::git2::Repository,
    pub config: Config,
}

impl Archive {
    /// Get the archive based on the CWD
    ///
    /// May fail if e.g. the archive does not exist.
    pub fn new() -> Result<Self, ArchiveError> {
        let path = Archive::get_root()?;

        let repo = git2::Repository::init(&path)?;
        let config = Archive::get_config_from_main_branch(&path)?;
        Ok(Self { path, repo, config })
    }

    /// Get all open imports
    pub fn get_imports(&self) -> Result<Vec<git2::Branch>, ArchiveError> {
        let branches = git::branch(&self.repo)?;
        Ok(branches
            .filter_map(|b| b.ok())
            .map(|(branch, _)| branch)
            .collect())
    }

    /// Switch to branch `branch_name`
    pub fn switch_branch(&self, branch_name: &String) -> Result<(), ArchiveError> {
        Ok(git::switch_branch(&self.repo, branch_name)?)
    }

    /// Get the name of the currently checkout branch.
    pub fn get_current_branch(&self) -> Result<String, ArchiveError> {
        Ok(git::get_current_branch(&self.repo)?)
    }

    /// Checks if a given branch exists locally.
    pub fn exists_branch(&self, branch_name: &String) -> bool {
        git::exists_branch(&self.repo, branch_name)
    }

    /// Indicate if the archive is in a dirty state.
    pub fn is_dirty(&self) -> Result<bool, ArchiveError> {
        Ok(git::is_dirty(&self.path)?)
    }

    /// Iteratively search archive root starting at CWD going up to /.
    ///
    /// Returns an error if CWD cannot be read or if not archive can be found.
    pub fn get_root() -> Result<PathBuf, ArchiveError> {
        let root = git::get_root()?;
        let mut annex_root = root.as_path().to_owned();
        annex_root.push(".git");
        annex_root.push("annex");

        if annex_root.exists() {
            Ok(root)
        } else {
            Err(ArchiveError::NoArchive)
        }
    }

    /// Get config by reading file.
    fn get_config_from_current_branch(&self) -> Result<Config, ArchiveError> {
        let config_path = Archive::get_root()?;
        let config_path = config_path.join(RELATIVE_CONFIG_FILE);
        Ok(Config::from_file(&config_path)?)
    }

    /// Get config by reading file from main branch.
    fn get_config_from_main_branch(path: &PathBuf) -> Result<Config, ArchiveError> {
        let main_branch = "main"; // TODO: How to figure out which main branch is, without config? :D
        let out = git::show(
            path,
            vec![format!("{}:{}", main_branch, RELATIVE_CONFIG_FILE)],
        )?;
        Ok(Config::from_string(&out)?)
    }
}
