mod activate;
mod deactivate;
mod list;
mod new;

use activate::activate;
use deactivate::deactivate;
use list::list;
use new::new;

use anyhow::{Context, Result};
use clap::{Args, Subcommand};
use std::path::{self, PathBuf};

use crate::archive::Archive;

pub fn run(import: Import) -> Result<()> {
    println!("Run import with {:?}", import);

    let archive = Archive::new().context("Failed to initiate archive")?;

    Ok(match import.command {
        ImportCommands::Activate { import_name } => activate(archive, import_name)?,
        ImportCommands::Add => todo!(),
        ImportCommands::Deactivate => deactivate(archive)?,
        ImportCommands::List => list(archive)?,
        ImportCommands::New {
            identifier,
            schema,
            images,
        } => new(archive, identifier, schema, images)?,
    })
}

#[derive(Args, Debug)]
pub struct Import {
    #[command(subcommand)]
    command: ImportCommands,
}

#[derive(Subcommand, Debug)]
enum ImportCommands {
    /// Activate an pending import
    Activate {
        /// Name of import to activate
        import_name: String,
    },

    /// Add images to the current import
    Add,

    /// Leave the activated import
    Deactivate,

    /// List all pending imports
    List,

    /// Create a new import
    New {
        /// Name describing the new import
        #[arg(short, long)]
        identifier: Option<String>,

        /// Schema to use for the new import
        #[arg(short, long)]
        schema: Option<String>,

        /// Images to import
        #[arg(required=true, num_args=1.., value_parser = validate_path)]
        images: Vec<PathBuf>,
    },
}

/// Ensures that a provided path is an existing file
fn validate_path(image: &str) -> Result<PathBuf, String> {
    let image: path::PathBuf = path::absolute(image).expect("Failed to get absolute path");
    if !image.exists() {
        Err(format!("File {} does not exist.", image.to_str().unwrap()))
    } else if !image.is_file() {
        Err(format!(
            "{} does not appear to be a file.",
            image.to_str().unwrap()
        ))
    } else {
        Ok(image)
    }
}
