/// Provides all functionality for creating a new import.
use anyhow::{bail, Result};
use chrono::prelude::NaiveDateTime;
use std::collections::HashSet;
use std::fs::{copy, create_dir_all};
use std::path::{Path, PathBuf};
use std::process::Command;
use std::thread;
use std::time::Duration;

use exif::Tag;

use crate::archive::Archive;
use crate::error::Metadata as MetadataError;
use crate::format_string::branch::BranchFS;
use crate::format_string::image::ImageFS;
use crate::git;
use crate::metadata::Metadata;

/// Represents a single image which is to be imported
struct ImportImage {
    /// Absolute path to the image
    pub source: PathBuf,
    /// Format string for the destination
    destination: ImageFS,
    /// Creation time
    datetime: NaiveDateTime,
    /// Metadata
    metadata: Metadata,
}

impl ImportImage {
    /// Initialize `ImportImage` by extracting the necessary information from the image.
    fn init(
        archive: &Archive,
        source: PathBuf,
        schema: &Option<String>,
        identifier: Option<String>,
    ) -> Result<Self, MetadataError> {
        debug_assert!(source.is_absolute());

        let metadata = Metadata::init(&source)?;
        let dt_raw = metadata.get_formatted(Tag::DateTimeOriginal)?;
        let dt = NaiveDateTime::parse_from_str(&dt_raw, "%Y-%m-%d %H:%M:%S")
            .map_err(|_| MetadataError::DateTimeInvalidFormat)?;

        let extension = source
            .extension()
            .and_then(|e| Some(String::from(e.to_str().unwrap())));

        Ok(Self {
            source,
            destination: ImageFS::init(&archive, schema, Some(dt), identifier, extension),
            datetime: dt,
            metadata,
        })
    }
}

/// Represents the state of a whole import
struct ImportState {
    /// List of images to import, sorted by creation time in ascending order (old to new).
    images: Vec<ImportImage>,
    /// Format string for the import branch
    branch: BranchFS,
}

impl ImportState {
    /// Initializes a new import using list of image paths.
    ///
    /// Iterates over all images, extracts the needed information, and stores them in the state.
    fn init(
        img_paths: Vec<PathBuf>,
        archive: &Archive,
        schema: Option<String>,
        identifier: Option<String>,
    ) -> Result<Self, MetadataError> {
        assert!(!img_paths.is_empty());

        let mut images: Vec<ImportImage> = Vec::new();
        for e in img_paths {
            debug_assert!(e.is_absolute());
            images.push(ImportImage::init(archive, e, &schema, identifier.clone())?);
        }

        // Sort images by capture time
        images.sort_by_key(|e| e.datetime);
        let start_dt = images.first().and_then(|e| Some(e.datetime));
        let end_dt = images.last().and_then(|e| Some(e.datetime));

        // Fix counts of images
        // Create set containing all definite destinations. Iterate through images oldest-to-newest
        // and check if the destination is already taken by checking if it is already in the set. If
        // not, insert the destination into the set. Otherwise, increment the count and repeat the
        // verification.
        let mut dest: HashSet<String> = HashSet::new();
        for e in images.iter() {
            while dest.contains(&e.destination.format()) {
                e.destination.state.increment_count();
            }
            dest.insert(e.destination.format());
        }
        assert!(dest.len() == images.len());

        Ok(ImportState {
            images,
            branch: BranchFS::init(&archive.config, start_dt, end_dt, identifier),
        })
    }
}

/// Create new import
pub fn new(
    archive: Archive,
    identifier: Option<String>,
    schema: Option<String>,
    images: Vec<PathBuf>,
) -> Result<()> {
    if archive.is_dirty()? {
        bail!("Archive is dirty. New import not possible");
    }

    println!(
        "Create new import. Identifier: {}, Schema {}",
        identifier
            .as_ref()
            .unwrap_or(&String::from("not specified")),
        schema.as_ref().unwrap_or(&String::from("not provided")),
    );

    let state: ImportState = ImportState::init(images, &archive, schema, identifier)?;

    prepare_branch_for_new_import(&archive, &state)?;

    // TODO: Actual import can be multi threaded as there (should be) no collisions in destination.

    // Copy image to destination
    // TODO: use git2 library
    for i in state.images.iter() {
        let formatted_dest = i.destination.format();
        let destination = Path::new(&formatted_dest);
        create_dir_all(
            destination
                .parent()
                .expect("There should always be a parent"),
        )?;
        copy(&i.source, destination)?;
        // TODO: somehow we get a permission error without timeout
        thread::sleep(Duration::from_millis(100));
    }

    // Add to git-annex
    // TODO: use batch option of git-annex
    for i in state.images.iter() {
        let formatted_dest = i.destination.format();
        Command::new("git-annex")
            .arg("add")
            .arg(formatted_dest)
            .output()?;
    }

    // TODO: use git2 library
    let args = vec![
        String::from("commit"),
        String::from("--amend"),
        String::from("--message"),
        format!("Import {} images", state.images.len()),
    ];
    git::run(&archive.path, args)?;

    println!("Done importing {} images", state.images.len());
    Ok(())
}

/// Ensure the import branch is unique, create it, and add an initial commit.
fn prepare_branch_for_new_import(archive: &Archive, state: &ImportState) -> Result<()> {
    // Ensure `branch_name` is unique by adding count if required
    let mut unique_branch = state.branch.format();
    while archive.exists_branch(&unique_branch) {
        println!("Branch {} exists already", unique_branch);
        state.branch.state.increment_count();
        unique_branch = state.branch.format();
    }

    println!("Found unique branch: {}", unique_branch);

    // TODO: Ensure the current branch is clean

    // Checkout new branch
    // TODO: use git2 library
    let args = vec![
        String::from("switch"),
        String::from("--orphan"),
        String::from(&unique_branch),
    ];
    git::run(&archive.path, args)?;

    // TODO: use git2 library
    let args = vec![String::from("clean"), String::from("--force")];
    git::run(&archive.path, args)?;

    // TODO: use git2 library
    let args = vec![
        String::from("commit"),
        String::from("--allow-empty"),
        String::from("--message"),
        String::from("NEW EMPTY IMPORT"),
    ];
    git::run(&archive.path, args)?;

    assert!(archive.get_current_branch().unwrap() == unique_branch);
    Ok(())
}
