//! Handle config

pub use ini::Ini;
use std::collections::HashMap;
use std::path::PathBuf;

use crate::error::{Config as ConfigError, ConfigFormat as ConfigFormatError};

pub static RELATIVE_CONFIG_FILE: &str = "mia.ini";

/// List all sections of the config
#[derive(Debug, PartialEq)]
pub struct Config {
    pub general: General,
    pub schema: Schema,
}

/// List all options of the 'General' section
#[derive(Debug, PartialEq)]
pub struct General {
    pub import_branch_prefix: String,
    pub import_branch_schema: String,
    pub main_branch: String,
}

/// List all options of the 'Schema' section
#[derive(Debug, PartialEq)]
pub struct Schema {
    pub default: String,
    pub custom: HashMap<String, String>,
}

impl Config {
    /// Initialize `Config` by reading it from a valid Ini config file at `config_path`.
    pub fn from_file(config_path: &PathBuf) -> Result<Self, ConfigError> {
        let ini_conf = Ini::load_from_file(&config_path).map_err(|source| match source {
            ini::Error::Io(e) => ConfigError::ReadFile { source: e },
            ini::Error::Parse(e) => ConfigError::from(e),
        })?;
        Config::from_ini(ini_conf)
    }

    /// Initialize `Config` from valid ini string `dump`.
    pub fn from_string(dump: &str) -> Result<Self, ConfigError> {
        let ini_conf = Ini::load_from_str(dump)?;
        Config::from_ini(ini_conf)
    }

    /// Initialize `Config` from a `Ini` config.
    fn from_ini(ini_conf: Ini) -> Result<Self, ConfigError> {
        let mut schemas = HashMap::from_iter(Self::read_section(&ini_conf, "Schema")?);
        let schema_default = schemas.remove("Default").ok_or(ConfigError::InvalidFormat(
            ConfigFormatError::Custom {
                message: format!("Field 'Default' is missing in section 'Schema'"),
            },
        ))?;

        Ok(Self {
            general: General {
                import_branch_prefix: Self::read_property(
                    &ini_conf,
                    "General",
                    "ImportBranchPrefix",
                )?,
                import_branch_schema: Self::read_property(
                    &ini_conf,
                    "General",
                    "ImportBranchSchema",
                )?,
                main_branch: Self::read_property(&ini_conf, "General", "MainBranch")?,
            },
            schema: Schema {
                default: schema_default,
                custom: schemas,
            },
        })
    }

    /// Dumps the default config to `file`.
    /// `file` must not exists.
    pub fn dump_default(file: &PathBuf) -> Result<(), ConfigError> {
        assert!(file.is_absolute());

        // TODO: Add flag to overwrite config in case it exists
        if file.exists() {
            return Err(ConfigError::WriteFile {
                source: std::io::Error::new(
                    std::io::ErrorKind::Other,
                    format!(": File {} already exists", file.display()),
                ),
            });
        }

        let def_config = Config::default();

        let mut conf = Ini::new();
        conf.with_section(Some("General"))
            .set(
                "ImportBranchPrefix",
                def_config.general.import_branch_prefix,
            )
            .set(
                "ImportBranchSchema",
                def_config.general.import_branch_schema,
            )
            .set("MainBranch", def_config.general.main_branch);

        conf.with_section(Some("Schema"))
            .set("Default", def_config.schema.default)
            .set(
                "Event",
                def_config
                    .schema
                    .custom
                    .get("Event")
                    .expect("Default trait is supposed to add an 'Event' schema"),
            );

        conf.write_to_file(file)
            .map_err(|source| ConfigError::WriteFile { source })?;

        Ok(())
    }

    /// Return value of key within section
    fn read_property(file: &Ini, section: &str, key: &str) -> Result<String, ConfigFormatError> {
        Ok(file
            .section(Some(section))
            .ok_or(ConfigFormatError::Custom {
                message: format!("Section '{}' is missing", section),
            })?
            .get(key)
            .ok_or(ConfigFormatError::Custom {
                message: format!("Field '{}' is missing in section '{}'", key, section),
            })?
            .to_string())
    }

    /// Return (key, value) pair for given section
    fn read_section(file: &Ini, section: &str) -> Result<Vec<(String, String)>, ConfigFormatError> {
        Ok(file
            .section(Some(section))
            .ok_or(ConfigFormatError::Custom {
                message: format!("Section '{}' is missing", section),
            })?
            .iter()
            .map(|(s, k)| (s.to_string(), k.to_string()))
            .collect())
    }
}

impl Default for Config {
    fn default() -> Self {
        Self {
            general: General::default(),
            schema: Schema::default(),
        }
    }
}

impl Default for General {
    fn default() -> Self {
        Self {
            import_branch_prefix: String::from("import/"),
            import_branch_schema: String::from("*Sy*Sm*Sd*i*c"),
            main_branch: String::from("main"),
        }
    }
}

impl Default for Schema {
    fn default() -> Self {
        let mut schema = HashMap::new();
        schema.insert(
            String::from("Event"),
            String::from("%Y/*Sy*Sm*Sd-*Ey*Em*Ed*i/%y%m%d/%y%m%d_%H%M%S*c.*x"),
        );
        Self {
            default: String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x"),
            custom: schema,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::conftest::*;
    use crate::error as Error;
    use pretty_assertions::assert_eq;
    use rstest::{fixture, rstest};
    use std::assert_matches::assert_matches;
    use std::fs;

    #[fixture]
    pub fn default_config_string() -> String {
        String::from(
            "[General]
ImportBranchPrefix=import/
ImportBranchSchema=*Sy*Sm*Sd*i*c
MainBranch=main

[Schema]
Default=%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x
Event=%Y/*Sy*Sm*Sd-*Ey*Em*Ed*i/%y%m%d/%y%m%d_%H%M%S*c.*x",
        )
    }

    #[fixture]
    pub fn default_general() -> General {
        General {
            import_branch_prefix: String::from("import/"),
            import_branch_schema: String::from("*Sy*Sm*Sd*i*c"),
            main_branch: String::from("main"),
        }
    }

    #[fixture]
    pub fn default_schema() -> Schema {
        let mut schema = HashMap::new();
        schema.insert(
            String::from("Event"),
            String::from("%Y/*Sy*Sm*Sd-*Ey*Em*Ed*i/%y%m%d/%y%m%d_%H%M%S*c.*x"),
        );
        Schema {
            default: String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x"),
            custom: schema,
        }
    }

    #[fixture]
    pub fn default_config(default_general: General, default_schema: Schema) -> Config {
        Config {
            general: default_general,
            schema: default_schema,
        }
    }

    #[fixture]
    pub fn default_config_ini(default_config_string: String) -> Ini {
        Ini::load_from_str(&default_config_string).unwrap()
    }

    #[rstest]
    fn general_default(default_general: General) {
        let general = General::default();
        assert_eq!(general, default_general);
    }

    #[rstest]
    fn schema_default(default_schema: Schema) {
        let schema = Schema::default();
        assert_eq!(schema, default_schema);
    }

    #[rstest]
    fn config_default(default_config: Config) {
        let config = Config::default();
        assert_eq!(config, default_config);
    }

    #[rstest]
    fn config_from_file(tmp_dir: PathBuf, default_config_string: String, default_config: Config) {
        // TODO: from_file also calls from_ini. Patch it?
        let mut path = tmp_dir;
        path.push("config");
        fs::write(&path, default_config_string).expect("Failed to write to file");

        let config = Config::from_file(&path).unwrap();
        assert_eq!(config, default_config);
    }

    #[test]
    fn config_from_file_invalid_path_inexisting() {
        // TODO: from_file also calls from_ini. Patch it?
        let mut path = PathBuf::new();
        path.push("INVALID");
        let config = Config::from_file(&path);
        assert_matches!(config, Err(Error::Config::ReadFile { source: _ }));
    }

    #[rstest]
    fn config_from_string(default_config_string: String, default_config: Config) {
        // TODO: from_string also calls from_ini. Patch it?
        let config = Config::from_string(&default_config_string).unwrap();
        assert_eq!(config, default_config);
    }

    #[rstest]
    #[case(String::from(""))]
    #[case(String::from("[General]"))]
    #[case(String::from(
        "[General]
ImportBranchPrefix=MISSING_FIELDS

[Schema]
Default=MORE_MISSING_FIELDS"
    ))]
    fn config_from_string_invalid(#[case] config_str: String) {
        // TODO: from_string also calls from_ini. Patch it?
        let config = Config::from_string(&config_str);
        assert!(config.is_err());
        assert!(matches!(config, Err(Error::Config::InvalidFormat(_))));
    }

    #[rstest]
    fn config_from_ini(default_config_ini: Ini, default_config: Config) {
        let config = Config::from_ini(default_config_ini).unwrap();
        assert_eq!(config, default_config);
    }

    #[rstest]
    #[case(Ini::load_from_str(&"").unwrap())]
    #[case(Ini::load_from_str(&"[General]").unwrap())]
    #[case(Ini::load_from_str(&"[General]
ImportBranchPrefix=MISSING_FIELDS

[Schema]
Default=MORE_MISSING_FIELDS").unwrap())]
    fn config_from_ini_invalid(#[case] config_ini: Ini) {
        let config = Config::from_ini(config_ini);
        assert!(matches!(config, Err(Error::Config::InvalidFormat(_))));
    }

    #[rstest]
    fn config_dump_default(tmp_dir: PathBuf, default_config_string: String) {
        let mut path = tmp_dir;
        path.push("dump");
        assert!(Config::dump_default(&path).is_ok());
        let content = fs::read_to_string(&path).expect("Failed to read from file");
        let content = content.trim();
        assert_eq!(content, default_config_string);
    }

    #[rstest]
    #[case(
        String::from("General"),
        String::from("ImportBranchPrefix"),
        String::from("import/")
    )]
    #[case(
        String::from("General"),
        String::from("ImportBranchSchema"),
        String::from("*Sy*Sm*Sd*i*c")
    )]
    #[case(
        String::from("General"),
        String::from("MainBranch"),
        String::from("main")
    )]
    #[case(
        String::from("Schema"),
        String::from("Default"),
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x")
    )]
    #[case(
        String::from("Schema"),
        String::from("Event"),
        String::from("%Y/*Sy*Sm*Sd-*Ey*Em*Ed*i/%y%m%d/%y%m%d_%H%M%S*c.*x")
    )]
    fn config_read_property(
        default_config_ini: Ini,
        #[case] section: String,
        #[case] key: String,
        #[case] expected: String,
    ) {
        let property = Config::read_property(&default_config_ini, &section, &key).unwrap();
        assert_eq!(property, expected);
    }

    #[rstest]
    fn config_read_property_invalid_section(default_config_ini: Ini) {
        let section = "INVALID_SECTION";
        let message: String = format!("Section '{}' is missing", section.to_owned());
        let property = Config::read_property(&default_config_ini, section, &"some_key");

        assert_matches!(property, Err(Error::ConfigFormat::Custom{message: m}) if m == message);
    }

    #[rstest]
    fn config_read_property_invalid_field(default_config_ini: Ini) {
        let section = "General";
        let key = "INVALID_KEY";
        let message: String = format!(
            "Field '{}' is missing in section '{}'",
            key.to_owned(),
            section.to_owned()
        );
        let property = Config::read_property(&default_config_ini, section, key);
        assert_matches!(property, Err(Error::ConfigFormat::Custom{message: m}) if m == message);
    }

    #[rstest]
    #[case(String::from("General"), vec![
(String::from("ImportBranchPrefix"), String::from("import/")),
(String::from("ImportBranchSchema"), String::from("*Sy*Sm*Sd*i*c")),
(String::from("MainBranch"), String::from("main"))])
    ]
    #[case(String::from("Schema"), vec![
(String::from("Default"), String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x")),
(String::from("Event"), String::from("%Y/*Sy*Sm*Sd-*Ey*Em*Ed*i/%y%m%d/%y%m%d_%H%M%S*c.*x")),
    ])]
    fn config_read_section(
        default_config_ini: Ini,
        #[case] section: String,
        #[case] expected: Vec<(String, String)>,
    ) {
        let section = Config::read_section(&default_config_ini, &section).unwrap();
        assert_eq!(section, expected);
    }

    #[rstest]
    fn config_read_section_invalid_section(default_config_ini: Ini) {
        let section = "INVALID_SECTION";
        let message: String = format!("Section '{}' is missing", section.to_owned());
        let property = Config::read_section(&default_config_ini, section);
        assert_matches!(property, Err(Error::ConfigFormat::Custom{message: m}) if m == message);
    }
}
