use thiserror::Error;

#[derive(Error, Debug)]
pub enum Archive {
    /// No archive could be found.
    #[error("No archive found")]
    NoArchive,
    /// Error coming from the underlying git library.
    #[error("General Git error")]
    Git(#[from] Git),
    /// Error coming from the config
    #[error("General Config error")]
    Config(#[from] Config),
}

impl From<git2::Error> for Archive {
    fn from(value: git2::Error) -> Self {
        Archive::Git(Git::Generic(value))
    }
}

/// Error related to Git
#[derive(Error, Debug)]
pub enum Git {
    /// Error related to IO
    #[error("IO Error: {0}")]
    Io(#[from] std::io::Error),
    /// Error coming from Git2
    /// TODO: split into more meaningful types
    #[error("Generic: {}", .0.message())]
    Generic(#[from] git2::Error),
    /// Head points to invalid reference for this operation
    #[error("Invalid reference: {0}")]
    InvalidReference(String),
}

/// Error related to the config
#[derive(Error, Debug)]
pub enum Config {
    /// Error reading from config
    #[error("Could not read from config file")]
    ReadFile { source: std::io::Error },
    /// Error writing to config
    #[error("Could not write to config file")]
    WriteFile { source: std::io::Error },
    /// Error parsing config due to invalid format
    #[error("Config has invalid format")]
    InvalidFormat(ConfigFormat),
}

impl From<ini::ParseError> for Config {
    fn from(value: ini::ParseError) -> Self {
        Self::InvalidFormat(ConfigFormat::Ini { source: value })
    }
}

impl From<ConfigFormat> for Config {
    fn from(value: ConfigFormat) -> Self {
        Self::InvalidFormat(value)
    }
}

/// Error related to an invalid config format
#[derive(Error, Debug)]
pub enum ConfigFormat {
    /// Format error detected by the own library
    #[error("{message}")]
    Custom { message: String },
    /// Format error detected by the underlying `ini` library
    #[error("{}", .source.msg)]
    Ini { source: ini::ParseError },
}

/// Errors related to metadata
#[derive(Error, Debug)]
pub enum Metadata {
    /// Error related to IO
    #[error("IO Error: {0}")]
    Io(#[from] std::io::Error),
    /// Cannot read metadata
    #[error("Cannot read metadata: {msg}")]
    InvalidMetadata { msg: String },
    /// Queries for invalid field
    #[error("Invalid field: {msg}")]
    InvalidField { msg: String },
    /// Other metadata errors from kamadak-exif
    #[error("Other exif errors")]
    ExifError { source: exif::Error },
    /// Extracted DateTime has invalid format
    #[error("Extracted DateTime has invalid format")]
    DateTimeInvalidFormat,
}

impl From<exif::Error> for Metadata {
    fn from(value: exif::Error) -> Self {
        match value {
            exif::Error::Io(source) => Metadata::Io(source),
            exif::Error::NotFound(msg) | exif::Error::InvalidFormat(msg) => {
                Metadata::InvalidMetadata {
                    msg: msg.to_string(),
                }
            }
            _ => Metadata::ExifError { source: value },
        }
    }
}
