pub mod branch;
/// Struct representing, and functions for working, with format strings.
pub mod image;

use chrono::prelude::NaiveDateTime;
use regex::{Captures, Regex};
use std::cell::Cell;

#[derive(Debug)]
pub struct FSState {
    image_dt: Option<NaiveDateTime>,
    start_dt: Option<NaiveDateTime>,
    end_dt: Option<NaiveDateTime>,
    identifier: Option<String>,
    extension: Option<String>,
    count: Cell<Option<usize>>,
}

impl FSState {
    pub fn init(
        image_dt: Option<NaiveDateTime>,
        start_dt: Option<NaiveDateTime>,
        end_dt: Option<NaiveDateTime>,
        identifier: Option<String>,
        extension: Option<String>,
        count: Option<usize>,
    ) -> Self {
        Self {
            count: Cell::new(count),
            image_dt,
            start_dt,
            end_dt,
            identifier,
            extension,
        }
    }

    /// Increment the count by +1
    ///
    /// If count is has not yet been set, initialize it to 1.
    ///
    /// Using Interior Mutability pattern to make this work.
    pub fn increment_count(&self) -> () {
        match self.count.get() {
            Some(e) => self.count.set(Some(e + 1)),
            None => self.count.set(Some(1)),
        }
    }
}

/// Format `input` according to ISO 8601.
///
/// If `None` is provided for `image_dt` the respective specifiers are **NOT** eaten up.
/// TODO: Eat specifiers
fn fmt_8601(input: &str, image_dt: &Option<NaiveDateTime>) -> String {
    if let Some(image_dt) = image_dt {
        return image_dt.format(input).to_string();
    }
    return String::from(input);
}

/// Format all static custom specifiers in `input`.
///
/// Supports:
/// - *i
/// - *x
/// - *X
/// - *S{8601}, where 8601 represents any 8601 specifier
/// - *E{8601}, where 8601 represents any 8601 specifier

/// Format custom specifier for identifier.
///
/// "*i" in `input` is replaced by "-identifier".
/// TODO: make exact formatting customizable
///
/// If `None` is provided for `identifier` the respective specifiers are eaten up.
fn fmt_identifier(input: &str, identifier: &Option<String>) -> String {
    let re = Regex::new(r"(\*i)").unwrap();
    return re
        .replace_all(input, |_: &Captures| match identifier {
            Some(e) => format!("-{}", e),
            None => String::from(""),
        })
        .into();
}

/// Format custom specifier for extension.
///
/// "*x" or "*X" in `input` is replaced by "extension" or "EXTENSION", respectively.
///
/// If `None` is provided for `extension` the respective specifiers are eaten up.
fn fmt_extension(input: &str, extension: &Option<String>) -> String {
    // Lowercase
    let re = Regex::new(r"(\*x)").unwrap();
    let out: String = re
        .replace_all(input, |_: &Captures| match extension {
            Some(e) => e.to_lowercase(),
            None => String::from(""),
        })
        .into();

    // Uppercase
    let re = Regex::new(r"(\*X)").unwrap();
    return re
        .replace_all(&out, |_: &Captures| match extension {
            Some(e) => e.to_uppercase(),
            None => String::from(""),
        })
        .into();
}

/// Format custom specifier for start/end datetime.
///
/// If `None` is provided for `start_dt` or `end_dt` the respective specifiers are eaten up.
fn fmt_start_end_datetime(
    input: &str,
    start_dt: &Option<NaiveDateTime>,
    end_dt: &Option<NaiveDateTime>,
) -> String {
    /// Format the custom regex `re` in string `inp` using date time `dt`.
    ///
    /// The regex must contain a matching group, which, when formatted using `format!("%{}",
    /// group)`, yields a valid ISO 8601 specifier.
    ///
    /// Is used to replace the start_date_time and end_state_time custom specifiers.
    fn format_custom_date_time(inp: &str, re: Regex, dt: &NaiveDateTime) -> String {
        re.replace_all(&inp, |caps: &Captures| {
            // Skip first element, which represents a full match.
            let c = caps.get(1).unwrap();
            let dt_specifier = format!("%{}", c.as_str());
            let dt_new = dt.format(&dt_specifier);
            dt_new.to_string()
        })
        .into()
    }

    let mut out = String::from(input);

    if let Some(start_dt) = start_dt {
        let re = Regex::new(r"\*S(\w{1})").unwrap();
        out = format_custom_date_time(&out, re, start_dt);
    } else {
        let re = Regex::new(r"(\*S\w{1})").unwrap();
        out = re.replace_all(&out, |_: &Captures| String::from("")).into();
    }

    if let Some(end_dt) = end_dt {
        let re = Regex::new(r"\*E(\w{1})").unwrap();
        out = format_custom_date_time(&out, re, end_dt);
    } else {
        let re = Regex::new(r"(\*E\w{1})").unwrap();
        out = re.replace_all(&out, |_: &Captures| String::from("")).into();
    }

    return out;
}

/// Format custom specifiers for count.
///
/// If `None` is provided for `count` the respective specifiers are eaten up.
fn fmt_count(input: &str, count: &Option<usize>) -> String {
    // TODO: Replace count specifiers
    let re = Regex::new(r"(\*c)").unwrap();
    return re
        .replace_all(&input, |_: &Captures| match count {
            Some(e) => format!("-{}", e),
            None => String::from(""),
        })
        .into();
}

#[cfg(test)]
mod tests {
    use super::*;
    use pretty_assertions::assert_eq;
    use rstest::{fixture, rstest};

    #[fixture]
    /// Fixture providing DateTime of *2022-01-02 03:04:05*
    fn datetime1() -> NaiveDateTime {
        NaiveDateTime::parse_from_str("2022-01-02 03:04:05", "%Y-%m-%d %H:%M:%S").unwrap()
    }

    #[fixture]
    /// Fixture providing DateTime of *2023-11-12 13:14:15*
    fn datetime2() -> NaiveDateTime {
        NaiveDateTime::parse_from_str("2023-11-12 13:14:15", "%Y-%m-%d %H:%M:%S").unwrap()
    }

    #[rstest]
    #[case(
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x"),
        String::from("2022/220102*i/220102_030405*c.*x")
    )]
    fn test_fmt_8601(datetime1: NaiveDateTime, #[case] raw: String, #[case] expected: String) {
        let out = fmt_8601(&raw, &Some(datetime1));
        assert_eq!(out, expected);
    }

    #[rstest]
    #[case(
        Some(String::from("test")),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed-test/%y%m%d_%H%M%S*c.*x")
    )]
    #[case(
        None,
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed/%y%m%d_%H%M%S*c.*x")
    )]
    #[case(
        Some(String::from("test")),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*i*c.*x"),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed-test/%y%m%d_%H%M%S-test*c.*x")
    )]
    fn test_fmt_identifier(
        #[case] identifier: Option<String>,
        #[case] raw: String,
        #[case] expected: String,
    ) {
        let out = fmt_identifier(&raw, &identifier);
        assert_eq!(out, expected);
    }

    #[rstest]
    #[case(
        Some(String::from("cr2")),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.cr2")
    )]
    #[case(
        Some(String::from("cr2")),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*X"),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.CR2")
    )]
    #[case(
        None,
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.")
    )]
    #[case(
        None,
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*X"),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.")
    )]
    fn test_fmt_extension_small(
        #[case] extension: Option<String>,
        #[case] raw: String,
        #[case] expected: String,
    ) {
        let out = fmt_extension(&raw, &extension);
        assert_eq!(out, expected);
    }

    #[rstest]
    #[case(
        Some(datetime1()),
        Some(datetime2()),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/220102_231112*i/%y%m%d_%H%M%S*c.*x")
    )]
    #[case(
        Some(datetime1()),
        Some(datetime2()),
        String::from("*Ey*Em*Ed*EH*EM*ES"),
        String::from("231112131415")
    )]
    #[case(
        Some(datetime1()),
        Some(datetime2()),
        String::from("%Y/*Sy*Sm*Sd*SH*SM*SS_*Ey*Em*Ed*EH*EM*ES*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/220102030405_231112131415*i/%y%m%d_%H%M%S*c.*x")
    )]
    #[case(
        Some(datetime1()),
        None,
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/220102_*i/%y%m%d_%H%M%S*c.*x")
    )]
    #[case(
        None,
        Some(datetime2()),
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/_231112*i/%y%m%d_%H%M%S*c.*x")
    )]
    #[case(
        None,
        None,
        String::from("%Y/*Sy*Sm*Sd_*Ey*Em*Ed*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/_*i/%y%m%d_%H%M%S*c.*x")
    )]
    fn test_fmt_start_end_datetime(
        #[case] start_dt: Option<NaiveDateTime>,
        #[case] end_dt: Option<NaiveDateTime>,
        #[case] raw: String,
        #[case] expected: String,
    ) {
        let out = fmt_start_end_datetime(&raw, &start_dt, &end_dt);
        assert_eq!(out, expected);
    }

    #[rstest]
    #[case(
        Some(1),
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S-1.*x")
    )]
    #[case(
        Some(10),
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S-10.*x")
    )]
    #[case(
        None,
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S*c.*x"),
        String::from("%Y/%y%m%d*i/%y%m%d_%H%M%S.*x")
    )]
    fn test_fmt_count(#[case] count: Option<usize>, #[case] raw: String, #[case] expected: String) {
        let out = fmt_count(&raw, &count);
        assert_eq!(out, expected);
    }
}
