/// Struct representing, and functions for working, with format strings.
use chrono::prelude::NaiveDateTime;

use super::{fmt_8601, fmt_count, fmt_extension, fmt_identifier, fmt_start_end_datetime, FSState};
use crate::archive::Archive;

/// Represents a format string.
/// A format string contains format specifiers, which are replaced during formatting, by concrete
/// values.
#[derive(Debug)]
pub struct ImageFS {
    raw: String,
    pub state: FSState,
}

impl ImageFS {
    /// Initialize `FormatString` representing the image destination.
    ///
    /// The destination schema is define by the optional argument, as:
    /// - If user provided a schema, verify that it is valid (e.g. listed in the config). If not
    ///   valid, print an error.
    /// - If no schema was provided, use a default schema.
    pub fn init(
        archive: &Archive,
        schema: &Option<String>,
        image_dt: Option<NaiveDateTime>,
        identifier: Option<String>,
        extension: Option<String>,
    ) -> Self {
        let schema: &String = schema
            .as_ref()
            .and_then(|e| {
                Some(
                    archive
                        .config
                        .schema
                        .custom
                        .get(e)
                        .unwrap_or_else(|| panic!("Schema {} is not valid.", e)),
                )
            })
            .unwrap_or(&archive.config.schema.default);

        Self {
            raw: schema.clone(),
            state: FSState::init(image_dt, None, None, identifier, extension, None),
        }
    }

    pub fn format(&self) -> String {
        let out = self.raw.clone();
        let out = fmt_8601(&out, &self.state.image_dt);
        let out = fmt_start_end_datetime(&out, &self.state.start_dt, &self.state.end_dt);
        let out = fmt_identifier(&out, &self.state.identifier);
        let out = fmt_extension(&out, &self.state.extension);
        let out = fmt_count(&out, &self.state.count.get());
        out
    }
}
