//! Provided git related functionalities
//! Wraps around git2

use std::env::current_dir;
use std::path::{Path, PathBuf};
use std::process::Command;

pub use git2;
use git2::{BranchType, Branches, Repository};

use crate::error::Git as GitError;

/// Iteratively search git repository starting at CWD going up to /.
///
/// Returns an error if CWD cannot be read or if not repository can be found.
pub fn get_root() -> Result<PathBuf, GitError> {
    let cwd = current_dir()?;
    let ceiling_dir = Path::new("/");
    // TODO: Improve error type of `discover_path` to indicate if no repo found or other error
    let repo = Repository::discover_path(cwd, ceiling_dir)?;
    let repo = repo.parent().expect("There should always be a parent");

    Ok(PathBuf::from(repo))
}

/// Convenience wrapper to switch to a specific branch.
///
/// Branch `branch_name` must exists.
pub fn switch_branch(repo: &Repository, branch_name: &str) -> Result<(), GitError> {
    // No need to change branch, if we are already on the right branch
    if let Ok(branch) = get_current_branch(repo) && &branch == branch_name {
        println!("Your are already on the right target. No need to switch.");
        return Ok(());
    }

    let reference = repo
        .find_branch(branch_name, BranchType::Local)?
        .into_reference();

    let object = reference.peel(git2::ObjectType::Any)?;

    repo.checkout_tree(&object, None)?;
    repo.set_head(
        reference
            .name()
            .expect("Branch name contains invalid utf-8 chars"),
    )?;

    println!("Switch to branch {}", branch_name);
    Ok(())
}

/// Get the name of the currently checkout branch.
pub fn get_current_branch(repo: &Repository) -> Result<String, GitError> {
    let head_ref = repo.head()?;

    if !head_ref.is_branch() {
        Err(GitError::InvalidReference(String::from(
            "Current reference is not a branch",
        )))
    } else {
        Ok(String::from(
            head_ref
                .shorthand()
                .expect("Branch name contains invalid utf-8 chars"),
        ))
    }
}

/// Git list of all local branches.
pub fn branch(repo: &Repository) -> Result<Branches, GitError> {
    Ok(repo.branches(Some(git2::BranchType::Local))?)
}

/// Checks if a given branch exists locally.
/// TODO: May return false even if branch exists
pub fn exists_branch(repo: &Repository, branch_name: &str) -> bool {
    repo.find_branch(branch_name, BranchType::Local).is_ok()
}

/// Indicate if the repository is dirty.
///
/// Dirty is interpreted as having uncommitted changes and/or having untracked files.
/// TODO: use git2 library
pub fn is_dirty(repository: &PathBuf) -> Result<bool, GitError> {
    let args = vec![String::from("status"), String::from("--porcelain")];
    let out = run(repository, args)?;
    Ok(!out.is_empty())
}

/// Run git command
pub fn run(repository: &PathBuf, args: Vec<String>) -> Result<String, GitError> {
    let out = Command::new("git")
        .args(&args)
        .current_dir(repository)
        .output()?;

    if out.status.success() {
        Ok(String::from_utf8_lossy(&out.stdout).trim().to_string())
    } else {
        Err(GitError::Io(std::io::Error::new(
            std::io::ErrorKind::Other,
            format!(
                "Failed to run `{}`",
                args.iter()
                    .fold(String::from("git"), |accum, e| format!("{} {}", accum, e))
            ),
        )))
    }
}

/// Run `git show ARGS` command.
///
/// TODO: use git2 library
pub fn show(repository: &PathBuf, args: Vec<String>) -> Result<String, GitError> {
    let mut args_complete = args.to_owned();
    args_complete.insert(0, String::from("show"));
    run(repository, args_complete)
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::conftest::tmp_dir;
    use pretty_assertions::assert_eq;
    use rstest::{fixture, rstest};
    use std::env::set_current_dir;
    use std::fs;

    /// Get name of current branch.
    fn get_current_branch2(repo: &PathBuf) -> String {
        let out = Command::new("git")
            .arg("rev-parse")
            .arg("--abbrev-ref")
            .arg("HEAD")
            .current_dir(repo)
            .output()
            .expect("Failed to run command");
        String::from_utf8_lossy(&out.stdout).trim().to_string()
    }

    /// Switch branch.
    fn switch_branch2(repo: &PathBuf, branch: &str) -> () {
        Command::new("git")
            .arg("switch")
            .arg(branch)
            .current_dir(repo)
            .output()
            .expect("Failed to run command");
    }

    /// Creates three branches.
    fn create_branches(repo: &PathBuf) -> () {
        Command::new("git")
            .arg("branch")
            .arg("branch1")
            .arg("main")
            .current_dir(&repo)
            .output()
            .expect("Failed to run command");
        Command::new("git")
            .arg("branch")
            .arg("branch2")
            .arg("main")
            .current_dir(&repo)
            .output()
            .expect("Failed to run command");
        Command::new("git")
            .arg("branch")
            .arg("branch3")
            .arg("main")
            .current_dir(&repo)
            .output()
            .expect("Failed to run command");
    }

    #[fixture]
    /// Initialize new git repository at temporary path.
    ///
    /// Changes working directory into the repo's root.
    fn repository(tmp_dir: PathBuf) -> (Repository, PathBuf) {
        let repo = Repository::init(&tmp_dir).unwrap();

        let _ = Command::new("git")
            .arg("commit")
            .arg("--allow-empty")
            .arg("--message")
            .arg("INITIAL COMMIT")
            .arg("--no-signoff")
            .arg("--no-gpg-sign")
            .current_dir(&tmp_dir)
            .output()
            .expect("Failed to run command");
        (repo, tmp_dir)
    }

    #[rstest]
    #[case(String::from(""))]
    #[case(String::from("2023"))]
    #[case(String::from("2023/230913"))]
    #[case(String::from(".git"))]
    fn test_get_root(repository: (Repository, PathBuf), #[case] relative_path: String) {
        let mut repo = repository.1;
        let repo_root = repo.to_owned();
        repo.push(relative_path);
        fs::create_dir_all(&repo).expect("Failed to create dir");
        set_current_dir(&repo).expect("Failed to change dir");
        let root = get_root().unwrap();
        assert_eq!(root, repo_root);
    }

    #[rstest]
    fn test_switch_branch(repository: (Repository, PathBuf)) {
        create_branches(&repository.1);

        assert_eq!(get_current_branch2(&repository.1), String::from("main"));

        assert!(switch_branch(&repository.0, "branch1").is_ok());
        assert_eq!(get_current_branch2(&repository.1), String::from("branch1"));

        assert!(switch_branch(&repository.0, "main").is_ok());
        assert_eq!(get_current_branch2(&repository.1), String::from("main"));

        assert!(switch_branch(&repository.0, "main").is_ok());
        assert_eq!(get_current_branch2(&repository.1), String::from("main"));

        assert!(switch_branch(&repository.0, "branch1").is_ok());
        assert_eq!(get_current_branch2(&repository.1), String::from("branch1"));

        assert!(switch_branch(&repository.0, "branch3").is_ok());
        assert_eq!(get_current_branch2(&repository.1), String::from("branch3"));
    }

    #[rstest]
    fn test_get_current_branch(repository: (Repository, PathBuf)) {
        create_branches(&repository.1);
        assert_eq!(
            get_current_branch(&repository.0).unwrap(),
            String::from("main")
        );

        switch_branch2(&repository.1, "branch1");
        assert_eq!(
            get_current_branch(&repository.0).unwrap(),
            String::from("branch1")
        );

        switch_branch2(&repository.1, "main");
        assert_eq!(
            get_current_branch(&repository.0).unwrap(),
            String::from("main")
        );

        switch_branch2(&repository.1, "branch1");
        assert_eq!(
            get_current_branch(&repository.0).unwrap(),
            String::from("branch1")
        );

        switch_branch2(&repository.1, "branch3");
        assert_eq!(
            get_current_branch(&repository.0).unwrap(),
            String::from("branch3")
        );
    }

    #[rstest]
    fn test_branch(repository: (Repository, PathBuf)) {
        create_branches(&repository.1);

        let mut branches = branch(&repository.0)
            .unwrap()
            .filter_map(|b| b.ok())
            .map(|(branch, _)| branch.name().unwrap().unwrap().to_string())
            .collect::<Vec<String>>();
        branches.sort();

        let expected = vec![
            String::from("branch1"),
            String::from("branch2"),
            String::from("branch3"),
            String::from("main"),
        ];

        assert_eq!(branches, expected);
    }

    #[rstest]
    #[case("main", true)]
    #[case("branch1", true)]
    #[case("branch3", true)]
    #[case("master", false)]
    #[case("invalid", false)]
    fn test_exists_branch(
        repository: (Repository, PathBuf),
        #[case] branch: &str,
        #[case] expected: bool,
    ) {
        create_branches(&repository.1);
        assert_eq!(exists_branch(&repository.0, branch), expected);
    }

    #[rstest]
    fn test_is_dirty(repository: (Repository, PathBuf)) {
        // Clean Repo
        assert!(!is_dirty(&repository.1).unwrap());

        // Untracked file
        let mut file = repository.1.to_owned();
        file.push("FILE1");
        fs::write(&file, "bla").unwrap();
        assert!(is_dirty(&repository.1).unwrap());

        // Staged file
        Command::new("git")
            .arg("add")
            .arg("FILE1")
            .current_dir(&repository.1)
            .output()
            .unwrap();
        assert!(is_dirty(&repository.1).unwrap());

        // Clean Repo
        Command::new("git")
            .arg("commit")
            .arg("--message")
            .arg("Add FILE1")
            .arg("--no-signoff")
            .arg("--no-gpg-sign")
            .current_dir(&repository.1)
            .output()
            .unwrap();
        assert!(!is_dirty(&repository.1).unwrap());

        // Unstaged Changes
        fs::write(&file, "new bla").unwrap();
        assert!(is_dirty(&repository.1).unwrap());

        // Staged changes
        Command::new("git")
            .arg("add")
            .arg("FILE1")
            .current_dir(&repository.1)
            .output()
            .unwrap();
        assert!(is_dirty(&repository.1).unwrap());

        // Clean Repo
        Command::new("git")
            .arg("commit")
            .arg("--message")
            .arg("Update FILE1")
            .arg("--no-signoff")
            .arg("--no-gpg-sign")
            .current_dir(&repository.1)
            .output()
            .unwrap();
        assert!(!is_dirty(&repository.1).unwrap());
    }

    // fn test_run() {}

    // fn test_show() {}
}
